from iosdebug.constants import RUN_SCRIPT_PHASE_ID
import os
import re


RUN_SCRIPT_PHASE = """
    BBF071DD2631DA3900DF13DA /* Sync debug mode */ = {
        isa = PBXShellScriptBuildPhase;
        buildActionMask = 2147483647;
        files = (
        );
        inputFileListPaths = (
        );
        inputPaths = (
        );
        name = "Sync debug mode";
        outputFileListPaths = (
        );
        outputPaths = (
        );
        runOnlyForDeploymentPostprocessing = 0;
        shellPath = /bin/sh;
        shellScript = "/usr/bin/python3 ${PROJECT_DIR}/.ios-debug_sync.py";
    };"""
BUILD_PHASE = "				BBF071DD2631DA3900DF13DA /* Sync debug mode */,"


RUN_SCRIPT_BUILD_PHASE = """
import os
from iosdebug.stop import stop
from iosdebug.start import start

if os.environ['ENABLE_PREVIEWS'] == "YES":
    raise SystemExit

stop(update_project=False)
start(update_project=False)
""".strip()


def add_run_script_build_phase(path):
    files = os.listdir(path)
    xcodeproj_file_name = [file for file in files if ".xcodeproj" in file][0]
    project_file = path + os.sep + xcodeproj_file_name + os.sep + "project.pbxproj"
    with open(project_file, "r") as file:
        content = file.read()
        if RUN_SCRIPT_PHASE_ID in content:
            return

        shell_scripts_part_end = "/* End PBXShellScriptBuildPhase section */"

        if shell_scripts_part_end in content:
            content = content.replace(
                shell_scripts_part_end, RUN_SCRIPT_PHASE + "\n" + shell_scripts_part_end
            )
        else:
            content += (
                shell_scripts_part_end.replace("End", "Start")
                + "\n"
                + RUN_SCRIPT_PHASE
                + "\n"
                + shell_scripts_part_end
            )

        native_target_section = re.findall(
            r"/\* Begin PBXNativeTarget section \*/([\s\S]+)/\* End PBXNativeTarget section \*/",
            content,
        )[0]
        targets = re.split(r"[A-Z0-9]{24} /\* .* \*/ =", native_target_section)[1:]
        app_flavours = [
            target
            for target in targets
            if 'productType = "com.apple.product-type.bundle.unit-test";' not in target
            and 'productType = "com.apple.product-type.bundle.ui-testing";'
            not in target
        ]
        build_phases_start = "buildPhases = ("
        for flavour in app_flavours:
            new_flavour = flavour.replace(
                build_phases_start, build_phases_start + "\n" + BUILD_PHASE
            )
            content = content.replace(flavour, new_flavour)

    with open(project_file, "w") as file:
        file.write(content)

    with open(".ios-debug_sync.py", "w") as file:
        file.write(RUN_SCRIPT_BUILD_PHASE)



import iosdebug.logger as logger
import re


def add_mock_registrations_to_container_builder(swift_files, registrations):
    print("Looking for ContainerBuilder.swift file")
    try:
        container_builder_file = [
            file for file in swift_files if "ContainerBuilder.swift" in file
        ][0]
    except IndexError:
        print("ContainerBuilder.swift file not found")
    with open(container_builder_file, "r") as file:
        content = file.read()
        registration_rows = (
            "// MARK: - DO NOT EDIT THIS PART WHILE IN DEBUG MODE ↓\n        "
            + "\n        ".join(registrations)
            + "\n        // MARK: - DO NOT EDIT THIS PART WHILE IN DEBUG MODE ↑"
        )
        repository_registration = re.findall(
            r"static func registerRepository[\s\S]*?\n    \}", content
        )[0]
        repository_registration_processed = repository_registration.replace(
            "return container", registration_rows + "\n        return container"
        )

    if registration_rows not in content:
        with open(container_builder_file, "w") as file:
            file.write(
                content.replace(
                    repository_registration, repository_registration_processed
                )
            )



import os


def find_swift_files(target_directory=os.getcwd()):
    swift_files = []

    for directory, _, files in os.walk(target_directory):
        for file in files:
            if (
                file[-6:] == ".swift"
                and ".build" not in directory
                and "Pods" not in directory
                and "Tests" not in directory
            ):
                swift_files.append(directory + os.sep + file)

    return swift_files



import re


def get_original_registrations(path_to_content_map, protocol_to_protocol_functions):
    files = list(path_to_content_map.values())
    result = {}
    for protocol in protocol_to_protocol_functions:
        for file in files:
            if "register(" + protocol in file:
                registration = re.findall(
                    r".*container.register\(" + protocol + r"[\s\S]+?}", file
                )[0].replace(protocol + ".self", protocol + "Impl.self")

                result[protocol] = registration
    return result



from iosdebug.processing.stopping.remove_run_script_build_phase import (
    remove_run_script_build_phase,
)
import os
from .processing.find_swift_files import find_swift_files
from .processing.create_path_to_content_dict import create_path_to_content_dict
from .processing.stopping import (
    store_mocked_implementations,
    set_original_root_view_controller,
    remove_mock_registrations,
)


def stop(path=os.getcwd(), update_project=True):
    swift_files = find_swift_files(path)
    path_to_content_map = create_path_to_content_dict(swift_files)
    store_mocked_implementations(path_to_content_map, path)
    set_original_root_view_controller(swift_files, path_to_content_map, path)
    remove_mock_registrations(swift_files)
    if update_project:
        remove_run_script_build_phase(path)



import os
import pickle
import re
from ast import literal_eval

from iosdebug.constants import DATA_FILE
from iosdebug.templates import SHAKABLE_NC_INSTANCE


def set_original_root_view_controller(swift_files, path_to_content_map, path):
    with open(path + os.sep + DATA_FILE, "rb") as file:
        data = pickle.load(file)
        data = literal_eval(data)

    for file_path in swift_files:
        with open(file_path, "r") as file:
            content = file.read()
            path_to_content_map[file_path] = content
            root_vc_property = re.findall(
                r"= (ShakableNavigationController.*)", content
            )
            if root_vc_property:
                content = content.replace(
                    root_vc_property[0], data["original_root_view_controller"]
                )
                content = content.split("\n// MARK: - Debug mode helper classes")[0]
                with open(file_path, "w") as file:
                    file.write(content)



import re
from iosdebug.processing.starting.create_basic_mock_implementation import (
    create_basic_mock_implementation,
)
from iosdebug.templates import FUNCTION_TEMPLATE


def get_updated_mocked_implementation(functions, mocked_impl, mock_functions):
    added_functions = []
    deleted_functions = []
    current_function_declarations = [func.declaration for func in functions]

    for function in functions:
        if function.declaration not in mocked_impl:
            added_functions.append(function)

    for function in mock_functions:
        if function not in current_function_declarations:
            deleted_functions.append(function)

    updated_mock_impl = mocked_impl

    for function in deleted_functions:
        function = function.replace("(", "\\(").replace(")", "\\)")
        updated_mock_impl = re.sub(
            r"[\s]+func " + function + r" {[\s\S]+?\n    }", "", updated_mock_impl
        )

    for function in added_functions:
        processed_func_template = create_basic_mock_implementation(function)
        extension_regex = r"extension .* {[\s\S]+?\n}"
        extension = re.findall(extension_regex, updated_mock_impl)[0]
        updated_extension = extension.replace(
            "\n}", "\n\n    " + processed_func_template + "\n}"
        )
        updated_mock_impl = re.sub(
            extension_regex, updated_extension, updated_mock_impl
        )

    return updated_mock_impl



import os
import pickle
from ast import literal_eval

from iosdebug.constants import DATA_FILE


def store_mocked_implementations(path_to_content_map, root_path):
    with open(root_path + os.sep + DATA_FILE, "rb") as file:
        data = pickle.load(file)
        data = literal_eval(data)
    for protocol in data["mock_implementations"]:
        for path in path_to_content_map:
            content = path_to_content_map[path]
            if "class Mock" + protocol in content:
                content_split = content.split("\n// MARK: - Mock implementation")
                if len(content_split) > 1:
                    data["mock_implementations"][protocol] = content_split[1]
                    with open(path, "w") as file:
                        file.write(content_split[0])
    with open(root_path + os.sep + DATA_FILE, "wb") as file:
        pickle.dump(str(data), file)



from iosdebug.templates import (
    MOCK_MANAGER_SETTINGS_TEMPLATE,
    FUNCTION_SETTINGS_TEMPLATE,
    MOCK_MANAGER_TEMPLATE,
)


def get_mock_settings_and_mock_cursor(
    protocol_to_protocol_functions, protocol_to_mock_function_variants
):
    mock_settings = []
    mock_cursors = []
    for repository in protocol_to_protocol_functions:
        funcs = protocol_to_protocol_functions[repository]
        func_annotations = []
        for func in funcs:
            if func.params:
                external_param_names = []
                for param in func.params:
                    external_param_name = param.split()[0]
                    if ":" not in external_param_name:
                        external_param_name += ":"
                    external_param_names.append(external_param_name)
            else:
                external_param_names = ""
            func_annotation = func.name + "(" + "".join(external_param_names) + ")"
            if func_annotation not in func_annotations:
                func_annotations.append(
                    func.name + "(" + "".join(external_param_names) + ")"
                )

        function_settings = []
        function_cursors = []
        for index, annotation in enumerate(func_annotations):
            func_declaration = funcs[index].declaration
            mocked_variants = protocol_to_mock_function_variants[repository][
                func_declaration
            ]
            function_settings.append(
                FUNCTION_SETTINGS_TEMPLATE.replace(
                    "<FUNCTION_ANNOTATION>",
                    annotation
                    # TODO: Implement getting mocked variants
                ).replace("<SETTINGS_VALUES>", "[" + ", ".join(mocked_variants) + "]")
            )
            function_cursors.append(
                FUNCTION_SETTINGS_TEMPLATE.replace(
                    "<FUNCTION_ANNOTATION>", annotation
                ).replace("<SETTINGS_VALUES>", "0")
            )

        processed_template = MOCK_MANAGER_SETTINGS_TEMPLATE.replace(
            "<PROTOCOL>", "Mock" + repository + "Impl"
        )
        processed_template_settings = processed_template.replace(
            "<FUNCTION_SETTINGS>", ",\n    ".join(function_settings)
        )
        processed_template_cursors = processed_template.replace(
            "<FUNCTION_SETTINGS>", ",\n    ".join(function_cursors)
        )

        mock_settings.append(processed_template_settings)
        mock_cursors.append(processed_template_cursors)

    processed_mock_manager = MOCK_MANAGER_TEMPLATE.replace(
        "<MOCK_CURSORS>", ",\n         ".join(mock_cursors)
    )
    processed_mock_manager = processed_mock_manager.replace(
        "<MOCK_SETTINGS>", ",\n         ".join(mock_settings)
    )
    return processed_mock_manager



from .find_repository_protocols import find_repository_protocols
from .write_mock_implementations import (
    write_mock_implementations,
)
from .add_mock_registrations_to_container_builder import (
    add_mock_registrations_to_container_builder,
)
from .get_protocols_to_protocol_functions_dict import (
    get_protocols_to_protocol_functions_dict,
)
from .get_mock_settings_and_mock_cursor import get_mock_settings_and_mock_cursor
from .get_protocol_to_mocked_content_dict import get_protocol_to_mocked_content_dict
from .store_mocked_implementation import store_mocked_implementation
from .create_and_set_mock_manager import create_and_set_mock_manager
from .get_registrations import get_registrations
from .get_mock_functions_variants import get_mock_functions_variants
from .is_debug_mode_on import is_debug_mode_on
from .add_run_script_build_phase import add_run_script_build_phase



import re


def remove_mock_registrations(swift_files):
    container_builder_file = [
        file for file in swift_files if "ContainerBuilder.swift" in file
    ][0]
    with open(container_builder_file, "r") as file:
        content = file.read()
        repository_registration = re.findall(
            r"static func registerRepository[\s\S]*?\n    \}", content
        )[0]
        repository_reg_rows = [
            row
            for row in repository_registration.split("\n")
            if "registerMock" not in row and "DO NOT EDIT THIS PART" not in row
        ]
        repository_registration_processed = "\n".join(repository_reg_rows)
    with open(container_builder_file, "w") as file:
        file.write(
            content.replace(repository_registration, repository_registration_processed)
        )



from iosdebug.processing.stopping.remove_run_script_build_phase import (
    remove_run_script_build_phase,
)
import os
import pickle

from .processing.find_swift_files import find_swift_files
from .processing.create_path_to_content_dict import create_path_to_content_dict
from .processing.starting import *
from iosdebug.constants import DATA_FILE, IGNORE_FILE


def start(path=os.getcwd(), update_project=True):
    if not os.path.isfile(path + os.sep + DATA_FILE):
        with open(path + os.sep + DATA_FILE, "wb") as file:
            pickle.dump("{'mock_implementations': {}}", file)

    if os.path.isfile(path + os.sep + IGNORE_FILE):
        with open(path + os.sep + IGNORE_FILE, "r") as file:
            ignored_protocols = file.read().split("\n")
    else:
        ignored_protocols = []

    swift_files = find_swift_files(path)

    if is_debug_mode_on(swift_files):
        print("Debug mode is already on.")
        return

    path_to_content_map = create_path_to_content_dict(swift_files)

    repository_protocols = find_repository_protocols(path_to_content_map)

    for protocol in ignored_protocols:
        if protocol in repository_protocols:
            index = repository_protocols.index(protocol)
            del repository_protocols[index]

    protocol_to_protocol_functions = get_protocols_to_protocol_functions_dict(
        repository_protocols, path_to_content_map
    )

    registrations = get_registrations(repository_protocols)

    write_mock_implementations(
        path_to_content_map, protocol_to_protocol_functions, path
    )

    protocol_to_mocked_contents = get_protocol_to_mocked_content_dict(
        repository_protocols, path_to_content_map
    )

    protocol_to_mock_function_variants = get_mock_functions_variants(
        protocol_to_mocked_contents
    )

    processed_mock_manager = get_mock_settings_and_mock_cursor(
        protocol_to_protocol_functions, protocol_to_mock_function_variants
    )

    store_mocked_implementation(protocol_to_mocked_contents, path)

    create_and_set_mock_manager(
        swift_files, path_to_content_map, processed_mock_manager, path
    )

    add_mock_registrations_to_container_builder(swift_files, registrations)

    if update_project:
        add_run_script_build_phase(path)



import re
from collections import namedtuple


def get_protocols_to_protocol_functions_dict(repository_protocols, path_to_content_map):
    Function = namedtuple(
        "Function", "declaration, name, generic_parameter_clause, params, return_type"
    )

    protocol_to_protocol_functions = {}
    for protocol in repository_protocols:
        for path in path_to_content_map:
            if "protocol " + protocol in path_to_content_map[path]:
                content = path_to_content_map[path]
                function_instances = []
                protocol_definition = re.findall(
                    "protocol " + protocol + r".* {([\s\S]*?)\n}", content
                )[0]
                protocol_definition_lines = [
                    line
                    for line in protocol_definition.split("\n")
                    if "//" not in line
                    and "var" not in line
                    and "typealias" not in line
                ]
                protocol_definition = "\n".join(protocol_definition_lines)
                functions = protocol_definition.split("func ")
                functions = [
                    func
                    for func in functions
                    if "typealias " not in func
                    and "var " not in func
                    and "MARK" not in func
                ]
                functions = [function.strip() for function in functions]
                functions = [function for function in functions if function]
                functions = [function.replace("\n", "") for function in functions]
                functions = [" ".join(function.split()) for function in functions]
                print("- " * 40)
                print("Found functions for protocol", protocol + ":")
                print()

                for func in functions:
                    function_name = re.findall(r"(\w+)", func)[0]
                    generic_parameter_clause_match = re.findall(
                        r"<.*>", func.split("(")[0]
                    )
                    if generic_parameter_clause_match:
                        generic_parameter_clause = generic_parameter_clause_match[0]
                    else:
                        generic_parameter_clause = None

                    params_part = func.replace(function_name, "", 1)
                    if "->" in func:
                        params_part = params_part.split("->")[0]
                    params_part = re.findall(r"\((.*)\)", params_part)[0]
                    if generic_parameter_clause:
                        params_part = params_part.replace(generic_parameter_clause, "")

                    params = params_part.split(", ")
                    for index, param in enumerate(params):
                        if "(" in param and not ")" in param:
                            params[index] = params[index] + ", " + params[index + 1]
                            del params[index + 1]
                    return_type = re.findall(r"-> (.*)", func)
                    if params[0] == "" or params[0] == "()":
                        params = None
                    if return_type:
                        return_type = return_type[0]
                    else:
                        return_type = None

                    print("func", func)
                    print("- name:", function_name)
                    print("- generic parameter clause:", generic_parameter_clause)
                    print("- params:", params)
                    print("- return type:", return_type)
                    function_instances.append(
                        Function(
                            func,
                            function_name,
                            generic_parameter_clause,
                            params,
                            return_type,
                        )
                    )
                    print()
                print("- " * 40)
                protocol_to_protocol_functions[protocol] = function_instances
                print()
    return protocol_to_protocol_functions



def create_path_to_content_dict(swift_files):
    path_to_content_map = {}

    for file_path in swift_files:
        with open(file_path, "r") as file:
            content = file.read()
            path_to_content_map[file_path] = content

    return path_to_content_map



import os
import pickle
import re
from ast import literal_eval

from iosdebug.constants import DATA_FILE
from iosdebug.templates import SHAKABLE_NC, SHAKABLE_NC_INSTANCE


def create_and_set_mock_manager(
    swift_files, path_to_content_map, processed_mock_manager, path
):
    for file_path in swift_files:
        with open(file_path, "r") as file:
            content = file.read()
            path_to_content_map[file_path] = content
            root_vc_property = re.findall(r"rootViewController[\s]+=[\s]+(.*)", content)

            if root_vc_property:
                if root_vc_property[0] == SHAKABLE_NC_INSTANCE:
                    break
                nav_con = "UINavigationController()"
                if root_vc_property[0] == nav_con:
                    root_vc_property_definition = nav_con
                    content = content.replace(nav_con, SHAKABLE_NC_INSTANCE)
                    changed = ""
                else:
                    root_vc_property_definition = re.findall(
                        "rootViewController = .*", content
                    )[0]
                    root_vc_property_instance = re.findall(
                        " = (.*)", root_vc_property_definition
                    )[0]
                    instance_definition = re.findall(
                        root_vc_property_instance + " = .*", content
                    )[0]
                    instance_assignment = re.findall(
                        root_vc_property_instance + " = (.*)", content
                    )[0]
                    changed = instance_definition.replace(
                        instance_assignment, SHAKABLE_NC_INSTANCE
                    )
                    content = content.replace(instance_definition, changed)
                with open(file_path, "w") as file:
                    file.write(content)
                    if "/ MARK: - Debug mode helper classes" not in content:
                        file.write(SHAKABLE_NC)
                        file.write(processed_mock_manager)

                with open(path + os.sep + DATA_FILE, "rb") as file:
                    data = pickle.load(file)
                    data = literal_eval(data)
                    data["original_root_view_controller"] = instance_assignment
                    data["changed_root_view_controlbler"] = changed

                with open(path + os.sep + DATA_FILE, "wb") as file:
                    pickle.dump(str(data), file)



import re
from pprint import pprint
from collections import defaultdict


def get_mock_functions_variants(protocol_to_mocked_contents):
    result = {}
    for entity in protocol_to_mocked_contents:
        result[entity] = defaultdict(str)
        mocked_content = protocol_to_mocked_contents[entity]
        mocked_extension = re.findall("extension [\s\S]+?\n}", mocked_content)[0]
        functions_split = mocked_extension.split("func ")[1:]
        for function in functions_split:
            function_name = re.findall("([\s\S]+?) {", function)[0]
            cases = re.findall("case (.*):", function)
            result[entity][function_name] = cases
    return result



from iosdebug.constants import RUN_SCRIPT_PHASE_ID
import re
import os


def remove_run_script_build_phase(path):
    files = os.listdir(path)
    xcodeproj_file_name = [file for file in files if ".xcodeproj" in file][0]
    project_file = path + os.sep + xcodeproj_file_name + os.sep + "project.pbxproj"
    with open(project_file, "r") as file:
        content = file.read()
        content = re.sub(
            ".*" + RUN_SCRIPT_PHASE_ID + " /\* Sync debug mode.+?,\n", "", content
        )
        content = re.sub(RUN_SCRIPT_PHASE_ID + r"[\s\S]+?}(;|,)\n", "", content)

    with open(project_file, "w") as file:
        file.write(content)



import os
import pickle

from iosdebug.constants import DATA_FILE


def store_mocked_implementation(protocol_to_mocked_contents, path):
    data = {"mock_implementations": protocol_to_mocked_contents}

    with open(path + os.sep + DATA_FILE, "wb") as file:
        pickle.dump(str(data), file)



import re


def find_repository_protocols(path_to_content_map):
    repository_protocols = []

    for file_path in path_to_content_map:
        content = path_to_content_map[file_path]
        repository_definition = re.findall(r"class .*Repository.*{[\s\S]*}", content)
        if repository_definition:
            protocol_conformance_row = re.findall(
                r"extension (\w+)" + "Impl: \\1 \{", repository_definition[0]
            )
            if protocol_conformance_row:
                repository_protocols.append(protocol_conformance_row[0])

    _ = [print(repo) for repo in repository_protocols]
    return repository_protocols



import sys
from collections import namedtuple
from iosdebug.start import start
from iosdebug.stop import stop
from prompt_toolkit.shortcuts import button_dialog, input_dialog, message_dialog
import iosdebug.logger as logger
import webbrowser


Command = namedtuple("Command", "name, description, execute")

commands = [
    Command("start", "Stars debug mode", start),
    Command("stop", "Stops debug mode", stop),
    Command(
        "issue",
        "Creates an issue",
        lambda: webbrowser.open("https://github.com/tomnvt/ios-debug/issues/new"),
    ),
    Command("exit", "Exits execution", lambda: sys.exit()),
]


def get_selected_command():
    arguments = sys.argv
    if len(arguments) < 2:
        print("No command found. Please use one of the following:")
        _ = [print(command.name, "-", command.description) for command in commands]
        sys.exit(0)
    return arguments[1]


def try_command_execution(selected_command):
    command_names = [command.name for command in commands]
    if selected_command in command_names:
        print("Executing command:", selected_command)
        index = command_names.index(selected_command)
        commands[index].execute()
    else:
        print("No command found. Please use one of the following:")
        _ = [print(command.name, "-", command.description) for command in commands]


def main():
    if len(sys.argv) == 1:
        command_index = button_dialog(
            title="iOS Debug",
            text="Select an action:",
            buttons=[
                (command.name[0].upper() + command.name[1:], index)
                for index, command in enumerate(commands)
            ],
        ).run()

    else:
        command = sys.argv[1]
        command_names = [command.name for command in commands]
        try:
            command_index = command_names.index(command)
        except ValueError:
            print("Command not found.")
            print("Options:\n")
            _ = [print(name) for name in command_names]

    commands[command_index].execute()
    logger.delete_old_log_file()


if __name__ == "__main__":
    main()



FUNCTION_TEMPLATE = """func <FUNC_NAME><GENERIC_PARAMETER_CLAUSE>(<FUNC_PARAMS>) <RETURN_TYPE>{
        switch MockManager.getOption(self, #function) {
        <CASES>
        default: break
        }
        <RETURN>originalInstance.<ORIGINAL_FUNC_CALL>
    }"""

TEMPLATE = """
// MARK: - Mock implementation
protocol Mock<PROTOCOL>: <PROTOCOL> {}

class Mock<PROTOCOL>Impl: <PROTOCOL> {

    private let originalInstance: <PROTOCOL>

    init(originalInstance: <PROTOCOL>) {
        self.originalInstance = originalInstance
    }
}

// MARK: - Mock protocol conformance
extension Mock<PROTOCOL>Impl {

    <FUNCTIONS>
}

import Swinject
func registerMock<PROTOCOL>(to container: Container) {
    // swiftlint:disable force_unwrapping
<ORIGINAL_REGISTRATION>
    container.register(<PROTOCOL>.self) { r in
        Mock<PROTOCOL>Impl(
            originalInstance: r.resolve(<PROTOCOL>Impl.self)!
        )
    }
}
"""

SHAKABLE_NC = """
// MARK: - Debug mode helper classes
// WARNING: DO NOT ADD ANY PRODUCTION CODE BELLOW THIS LINE
class ShakableNavigationController: UINavigationController {

    override func motionEnded(_ motion: UIEvent.EventSubtype, with event: UIEvent?) {
        if motion == .motionShake {
            present(MockManager(), animated: true, completion: nil)
        }
    }
}
"""

SHAKABLE_NC_INSTANCE = (
    "ShakableNavigationController() // !!! Don't edit this line while in debug mode !!!"
)

MOCK_MANAGER_TEMPLATE = """
// swiftlint:disable duplicate_imports
import Stevia

class MockManager: UITableViewController {

    static var cursor = [
        <MOCK_CURSORS>
    ]

    static var settings = [
        <MOCK_SETTINGS>
    ]

    var hiddenSections = Set<Int>()

    init() {
        super.init(nibName: nil, bundle: nil)
        hiddenSections = Set(0..<Self.settings.keys.count)
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
}

// MARK: - Boilerplate
// swiftlint:disable type_contents_order
extension MockManager {

    override func tableView(_ tableView: UITableView, titleForHeaderInSection section: Int) -> String? {
        let keys = Array(Self.settings.keys)
        return keys[section]
    }

    override func numberOfSections(in tableView: UITableView) -> Int {
        Self.settings.keys.count
    }

    // swiftlint:disable force_unwrapping
    override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        if self.hiddenSections.contains(section) {
            return 0
        }

        let keys = Array(Self.settings.keys)
        return Self.settings[keys[section]]?.keys.count ?? 0
    }

    override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = UITableViewCell()
        let currentKeySubkeyPair = getCurrentKeyAndSubKey(for: indexPath)
        let currentCursor = getCurrentCursor(for: indexPath)
        let functionVariant = Self.settings[currentKeySubkeyPair.key]![currentKeySubkeyPair.subKey]!
        cell.textLabel?.text = currentKeySubkeyPair.subKey + " - " + functionVariant[currentCursor]
        cell.textLabel?.numberOfLines = 0
        cell.textLabel?.lineBreakMode = .byWordWrapping
        return cell
    }

    override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        let currentKeySubkeyPair = getCurrentKeyAndSubKey(for: indexPath)
        let options = Self.settings[currentKeySubkeyPair.key]![currentKeySubkeyPair.subKey]!
        present(
            SelectVariantViewController(
                header: Array(Self.settings[currentKeySubkeyPair.key]!.keys)[indexPath.row],
                options: options,
                onResult: { index in
                    Self.cursor[currentKeySubkeyPair.key]![currentKeySubkeyPair.subKey] = index
                    tableView.reloadData()
                }),
            animated: true,
            completion: nil
        )
    }

    func getCurrentCursor(for indexPath: IndexPath) -> Int {
        let keyAndSubkey = getCurrentKeyAndSubKey(for: indexPath)
        return Self.cursor[keyAndSubkey.key]![keyAndSubkey.subKey]!
    }

    func getCurrentKeyAndSubKey(for indexPath: IndexPath) -> (key: String, subKey: String) {
        let keys = Array(Self.settings.keys)
        let currentKey = keys[indexPath.section]
        return (key: currentKey, subKey: Array(Self.settings[currentKey]!.keys)[indexPath.row])
    }

    static func getOption(_ target: Any, _ functionName: String) -> String {
        let className = Mirror(reflecting: target).description.split(separator: " ").last!
        let cursor = MockManager.cursor[String(className)]![functionName]!
        let options = MockManager.settings[String(className)]![functionName]!
        return options[cursor]
    }

    override func tableView(_ tableView: UITableView, viewForHeaderInSection section: Int) -> UIView? {
        let title = Array(Self.settings.keys)[section]
            .replacingOccurrences(of: "Mock", with: "")
            .replacingOccurrences(of: "Impl", with: "")
        let view = UIView()

        let btnSection = UIButton()
        btnSection.contentEdgeInsets = UIEdgeInsets(top: 20, left: 15, bottom: 20, right: 15)
        btnSection.setTitle(title,
                            for: .normal)
        btnSection.sizeToFit()
        btnSection.titleLabel?.numberOfLines = 0
        btnSection.tag = section
        btnSection.addTarget(self,
                             action: #selector(self.hideSection(sender:)),
                             for: .touchUpInside)
        btnSection.backgroundColor = UIColor.black.withAlphaComponent(0.1)
        btnSection.setTitleColor(.black, for: .normal)

        let viewSeparator = UIView()
        viewSeparator.backgroundColor = .black
        viewSeparator.height(1)

        view.subviews(btnSection, viewSeparator)
        view.layout(0,
                    |btnSection|,
                    0,
                    viewSeparator,
                    0)

        return view
    }

    @objc
    private func hideSection(sender: UIButton) {
        let section = sender.tag

        func indexPathsForSection() -> [IndexPath] {
            let keys = Array(Self.settings.keys)
            let optionsCount = Self.settings[keys[section]]!.count

            var indexPaths = [IndexPath]()

            for row in 0..<optionsCount {
                indexPaths.append(IndexPath(row: row,
                                            section: section))
            }

            return indexPaths
        }

        if self.hiddenSections.contains(section) {
            self.hiddenSections.remove(section)
            self.tableView.insertRows(at: indexPathsForSection(),
                                      with: .fade)
        } else {
            self.hiddenSections.insert(section)
            self.tableView.deleteRows(at: indexPathsForSection(),
                                      with: .fade)
        }
    }
}

class SelectVariantViewController: UITableViewController {

    private let header: String
    private let options: [String]
    private let onResult: (Int) -> Void

    init(header: String, options: [String], onResult: @escaping (Int) -> Void) {
        self.header = header
        self.options = options
        self.onResult = onResult
        super.init(nibName: nil, bundle: nil)
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func numberOfSections(in tableView: UITableView) -> Int {
        1
    }

    override func tableView(_ tableView: UITableView, titleForHeaderInSection section: Int) -> String? {
        header
    }

    override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return options.count
    }

    override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = UITableViewCell()
        cell.textLabel?.text = options[indexPath.row]
        return cell
    }

    override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        onResult(indexPath.row)
        dismiss(animated: true, completion: nil)
    }
}
"""

MOCK_MANAGER_SETTINGS_TEMPLATE = """"<PROTOCOL>": [
    <FUNCTION_SETTINGS>
        ]"""

FUNCTION_SETTINGS_TEMPLATE = """        "<FUNCTION_ANNOTATION>": <SETTINGS_VALUES>"""



def get_registrations(repository_protocols):
    registrations = []
    cases = ["Mocked1", "Mocked2", "Mocked3"]
    for protocol in repository_protocols:
        registrations.append(
            "registerMock<PROTOCOL>(to: container)".replace("<PROTOCOL>", protocol)
        )
    return registrations



import re


def is_debug_mode_on(swift_files):
    for file_path in swift_files:
        with open(file_path, "r") as file:
            content = file.read()
            if "ShakableNavigationController" in content:
                return True

    return False



import os
import pickle
from ast import literal_eval

from iosdebug.constants import DATA_FILE, KEY_MOCK_IMPLEMENTATIONS


def get_stored_mock_implementation(path, repository_name):
    with open(path + os.sep + DATA_FILE, "rb") as file:
        content = pickle.load(file)
        dictionary = literal_eval(content)
        mock_implementations = dictionary[KEY_MOCK_IMPLEMENTATIONS]
        if repository_name in mock_implementations:
            return mock_implementations[repository_name]



from .remove_mock_registrations import remove_mock_registrations
from .set_original_root_view_controller import set_original_root_view_controller
from .store_mocked_implementations import store_mocked_implementations



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="ios-debug",
    version="0.1.2",
    description="Debug mode manager for iOS.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="",
    author="Tom Novotny",
    author_email="tom.novota@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(include=["iosdebug", "iosdebug.*"]),
    include_package_data=True,
    install_requires=["prompt-toolkit", "psutil", "gitpython"],
    entry_points={
        "console_scripts": [
            "ios-debug=iosdebug.__main__:main",
        ]
    },
)



import logging
import os

LOG_FILE_NAME = "ios-debug.log"

instance = logging.getLogger(__name__)

# Create handlers
c_handler = logging.StreamHandler()
f_handler = logging.FileHandler(LOG_FILE_NAME)
c_handler.setLevel(logging.WARNING)
f_handler.setLevel(logging.WARNING)

# Create formatters and add it to handlers
c_format = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
f_format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
c_handler.setFormatter(c_format)
f_handler.setFormatter(f_format)

# Add handlers to the logger
instance.addHandler(c_handler)
instance.addHandler(f_handler)


def delete_old_log_file():
    os.remove(LOG_FILE_NAME)



def get_protocol_to_mocked_content_dict(repository_protocols, path_to_content_map):
    protocol_to_mocked_contents = {}

    for protocol in repository_protocols:
        for file_path in path_to_content_map:
            if "protocol " + protocol in path_to_content_map[file_path]:
                with open(file_path, "r") as file:
                    content = file.read()
                content_split = content.split("// MARK: - Mock implementation")
                mocked_content = content_split[1][1:]
                protocol_to_mocked_contents[protocol] = mocked_content
    return protocol_to_mocked_contents



from iosdebug.processing.starting.get_mock_functions_variants import (
    get_mock_functions_variants,
)
from iosdebug.processing.starting.get_updated_mocked_implementation import (
    get_updated_mocked_implementation,
)
from iosdebug.processing.starting.create_basic_mock_implementation import (
    create_basic_mock_implementation,
)
from iosdebug.processing.starting.get_original_registration import (
    get_original_registrations,
)
from iosdebug.processing.starting.get_stored_mock_implementation import (
    get_stored_mock_implementation,
)
from iosdebug.templates import FUNCTION_TEMPLATE, TEMPLATE


def write_mock_implementations(
    path_to_content_map, protocol_to_protocol_functions, root_path
):
    original_registrations = get_original_registrations(
        path_to_content_map, protocol_to_protocol_functions
    )
    for protocol in protocol_to_protocol_functions:
        for path in path_to_content_map:
            if "protocol " + protocol in path_to_content_map[path]:
                stored_impl = get_stored_mock_implementation(root_path, protocol)
                if stored_impl:
                    mock_variants = get_mock_functions_variants({protocol: stored_impl})
                    updated_impl = get_updated_mocked_implementation(
                        protocol_to_protocol_functions[protocol],
                        stored_impl,
                        mock_variants[protocol],
                    )
                    processed_template = (
                        "\n// MARK: - Mock implementation" + updated_impl
                    )
                else:
                    functions = []
                    for function in protocol_to_protocol_functions[protocol]:
                        processed_func_template = create_basic_mock_implementation(
                            function
                        )
                        functions.append(processed_func_template)

                    processed_template = TEMPLATE
                    processed_template = processed_template.replace(
                        "<PROTOCOL>", protocol
                    )

                    processed_template = processed_template.replace(
                        "<FUNCTIONS>", "\n\n    ".join(functions)
                    )

                try:
                    original_registration = original_registrations[protocol]
                    original_registration_initial_spaces = len(
                        original_registration
                    ) - len(original_registration.lstrip())
                    indentation_difference = original_registration_initial_spaces - 4
                    lines = original_registration.split("\n")
                    matched_indent = [line[indentation_difference:] for line in lines]
                    original_registration = "\n".join(matched_indent)

                    processed_template = processed_template.replace(
                        "<ORIGINAL_REGISTRATION>", original_registration
                    )
                except KeyError:
                    print("No registration found for protoco " + protocol)
                    processed_template = processed_template.replace(
                        "<ORIGINAL_REGISTRATION>",
                        "// !!! No original registration found !!!",
                    )

                with open(path, "a") as file:
                    file.write(processed_template)



from iosdebug.templates import FUNCTION_TEMPLATE


def create_basic_mock_implementation(function):
    processed_func_template = FUNCTION_TEMPLATE
    processed_func_template = processed_func_template.replace(
        "<FUNC_NAME>", function.name
    )
    if function.generic_parameter_clause:
        generic_param = function.generic_parameter_clause
    else:
        generic_param = ""
    processed_func_template = processed_func_template.replace(
        "<GENERIC_PARAMETER_CLAUSE>", generic_param
    )
    if function.params:
        processed_func_template = processed_func_template.replace(
            "<FUNC_PARAMS>", ", ".join(function.params)
        )
    else:
        processed_func_template = processed_func_template.replace("<FUNC_PARAMS>", "")

    cases = ["Original", "Mocked1", "Mocked2", "Mocked3"]
    processed_func_template = processed_func_template.replace(
        "<CASES>",
        "\n        ".join(['case "' + case + '": break' for case in cases]),
    )
    if function.params:
        split_parameters = [param.split(":")[0].split() for param in function.params]
        arguments = []
        for param_names in split_parameters:
            if len(param_names) == 2:
                if param_names[0] == "_":
                    arguments.append(param_names[1])
                else:
                    arguments.append(param_names[0] + ": " + param_names[1])
            else:
                arguments.append(param_names[0] + ": " + param_names[0])
        func_call = "(" + ", ".join(arguments) + ")"
    else:
        func_call = "()"
    processed_func_template = processed_func_template.replace(
        "<ORIGINAL_FUNC_CALL>", function.name + func_call
    )
    if function.return_type:
        processed_func_template = processed_func_template.replace(
            "<RETURN_TYPE>", "-> " + function.return_type + " "
        )
        processed_func_template = processed_func_template.replace("<RETURN>", "return ")
    else:
        processed_func_template = processed_func_template.replace("<RETURN_TYPE>", "")
        processed_func_template = processed_func_template.replace("<RETURN>", "")

    return processed_func_template
